#!/usr/bin/env bash

###############################################################################
# Copyright (c) 2022 NVIDIA Corporation. All rights reserved.
#
# This Software is licensed under one of the following licenses:
#
# 1) under the terms of the "Common Public License 1.0" a copy of which is
#    available from the Open Source Initiative, see
#    http://www.opensource.org/licenses/cpl.php.
#
# 2) under the terms of the "The BSD License" a copy of which is
#    available from the Open Source Initiative, see
#    http://www.opensource.org/licenses/bsd-license.php.
#
# 3) under the terms of the "GNU General Public License (GPL) Version 2" a
#    copy of which is available from the Open Source Initiative, see
#    http://www.opensource.org/licenses/gpl-license.php.
#
# Licensee has the right to choose one of the above licenses.
#
# Redistributions of source code must retain the above copyright
# notice and one of the license notices.
#
# Redistributions in binary form must reproduce both the above copyright
# notice, one of the license notices in the documentation
# and/or other materials provided with the distribution.
###############################################################################

set -Eeu
trap cleanup SIGHUP SIGINT SIGTERM ERR
SCRIPT_DIR=$(cd "$(dirname "${BASH_SOURCE[0]}")" &>/dev/null && pwd -P)
LOG_FILE="/var/log/sfc-install.log"

cleanup() {
  die "Installation failed. Cleaning up"
}


## Global variables
BR_HBN_NAME=${BR_HBN_NAME:-br-hbn}
MLNX_SF_FILE="/etc/mellanox/mlnx-sf.conf"
MLNX_OVS_FILE="/etc/mellanox/mlnx-ovs.conf"
MLNX_BF_FILE="/etc/mellanox/mlnx-bf.conf"
HBN_CONF_FILE="/etc/mellanox/hbn.conf"
SFC_CONF_FILE="/etc/mellanox/sfc.conf"
UDEV_SCRIPT_PATH="/usr/lib/udev"
SF_RENAME_SCRIPT="auxdev-sf-netdev-rename"
SF_REP_RENAME_SCRIPT="sf-rep-netdev-rename"
OVS_MAPPING_TEMPLATE_FILE="sfc.conf.tmpl"
IFUP_HOOK_50="50-ifup-hooks"
IFUP_HOOKS_PATH="/etc/networkd-dispatcher/routable.d"
CNI_CONF="10-containerd-net-br-mgmt.conflist"
CNI_CONF_PATH="/etc/cni/net.d/"
SYSCTL_CONF="50-sfc-hbn.conf"
SYSCTL_PATH="/lib/sysctl.d"
REBOOT_OPTION="false"
USER_PCI_ADDR=""
ECPF0=""
ECPF1=""
MIN_VF=0
MAX_VF=127
CPU_AFFINITY_LIST="0-2"
CONFIG_OPTIONS=""
DEBUG_OPTION=""
CLOUD_OPTION=${CLOUD_OPTION:-""}
ENABLE_BR_SFC=${ENABLE_BR_SFC:-""}
ENABLE_BR_SFC_DEFAULT_FLOWS=${ENABLE_BR_SFC_DEFAULT_FLOWS:-""}

# Host/Network facing SFs base IDs
BASE_NET_FACING_SF_NUM=2   # 0x1 - 0xE38 (2 - 1000)
BASE_HOST_FACING_SF_NUM=1001   # 0x3E9 - 0x7D0 (1001 - 2000)
BASE_DPU_FACING_SF_NUM=500
# Host facing SFs distributed in the following way:
# 0x3E9..0x438 - SFs mapped to ECPF0 VFs (pf0vf*_sf)  sfnum range: 1001 - 1256
ECPF0_VF_SF_BASE_ID=$BASE_HOST_FACING_SF_NUM  # sfnum: 1001
# 0x439..0x5E8 - SFs mapped to ECPF1 VFs (pf1vf*_sf)  sfnum range: 1257 - 1512
ECPF1_VF_SF_BASE_ID=$((BASE_HOST_FACING_SF_NUM + 0x100)) # sfnum: 1257
#DPU SFs - starting from 500
# 0x5EA,0x5EB - SFs mapped to Host PF representors (pf0hpf_sf, pf1hpf_sf)
PF0HPF_SF_ID=$((BASE_HOST_FACING_SF_NUM + 0x201)) # sfnum: 1514
PF1HPF_SF_ID=$((BASE_HOST_FACING_SF_NUM + 0x202)) # sfnum: 1515

# NVCONFIG variables corresponding to the default number of 16 SFs
PF_BAR2_ENABLE=${PF_BAR2_ENABLE:-0}
PER_PF_NUM_SF=${PER_PF_NUM_SF:-1}
PF_TOTAL_SF=${PF_TOTAL_SF:-40}
PF_SF_BAR_SIZE=${PF_SF_BAR_SIZE:-10}
NUM_PF_MSIX_VALID=${NUM_PF_MSIX_VALID:-0}
PF_NUM_PF_MSIX_VALID=${PF_NUM_PF_MSIX_VALID:-1}
PF_NUM_PF_MSIX=${PF_NUM_PF_MSIX:-228}

distro=`lsb_release -i -s``lsb_release -r -s | tr -d '.'`
distro=${distro,,}

if [ "$EUID" -ne 0 ]
  then echo "Please run as root"
  exit
fi

usage() {
  cat << EOF
Usage: $(basename "${BASH_SOURCE[0]}") [-p0 | --ecpf0 <num>]  [-p1 | --ecpf1 <num>] [-h] [-v] [-r] [-d] [-hs | --hugepagesz] [-hc | --hugepages]
Generate SF configs according to the number of VFs created on both physical ports
and install OVS SFC configuration for HBN.

Available options:

-p0, --ecpf0         [Optional] Number of VFs created on ECPF0 in range $MIN_VF..$MAX_VF (default: 0)
-p1, --ecpf1         [Optional] Number of VFs created on ECPF1 in range $MIN_VF..$MAX_VF (default: 0)
-r,  --reboot        [Optional] Reboot after install is completed
-d,  --debug         [Optional] Set vswitchd common options in DBG mode
-hs,  --hugepagesz   [Optional] Set hugepagesz size assumed to be in kB (default: 2048)
-hc,  --hugepages    [Optional] Set hugepages number (default: 3072)
-h,  --help          [Optional] Print this help and exit
-v,  --verbose       [Optional] Print script debug info
-c,  --config-file   [Optional] Use config ini file

EOF
  exit
}

TIMESTAMP=`date "+%Y-%m-%d %H:%M:%S"`
msg() {
  echo 2>&1 -e "[$TIMESTAMP]   ${1-}" | tee -ia ${LOG_FILE}
}

die() {
  local msg=$1
  local code=${2-1} # default exit status 1
  msg "Error: $msg"
  exit "$code"
}

is_number(){
  local string=$1
  case $string in
    ''|*[!0-9]*) die "$string : Not a valid number" ;;
    *) return ;;
  esac
}

validate_address(){
  pci_address_regex="^[a-fA-F0-9]{4}:[a-fA-F0-9]{2}\:[a-fA-F0-9]{2}\.[a-fA-F0-9]{1}$"
  [[ "$USER_PCI_ADDR" =~  $pci_address_regex ]] && msg "matched" || die "Bad pci address format"
}

validate_vf_number(){
  local ecpf=$1
  local num=$2
  is_number $num
  if (($num < $MIN_VF || $num > $MAX_VF)); then
    die "Invalid number of $ecpf VFs: $num. Please enter number in range $MIN_VF..$MAX_VF"
  fi
}

parse_params() {
  # Local variables, if any

  while :; do
    case "${1-}" in
    -h | --help) usage ;;
    -v | --verbose) set -x ;;
    -p0 | --ecpf0)
      ECPF0="${2-}"
      validate_vf_number ECPF0 $ECPF0
      shift
      ;;
    -p1 | --ecpf1)
      ECPF1="${2-}"
      validate_vf_number ECPF1 $ECPF1
      shift
      ;;
    -r | --reboot)  REBOOT_OPTION="true" ;;
    -d | --debug)  DEBUG_OPTION="true" ;;
    -c | --config-file)  CONFIG_OPTIONS="true";;
    -hs | --hugepagesz)
      HUGEPAGESZ="${2-}"
      shift
      ;;
    -hc | --hugepages)
      HUGEPAGES="${2-}"
      shift
      ;;
    -?*) die "Unknown option: $1" ;;
    *) break ;;
    esac
    shift
  done

  return 0
}

parse_params "$@"

function sf_create() {
  sfaddcmd="/sbin/mlnx-sf --action create $@ -t --cpu-list $CPU_AFFINITY_LIST"
  msg "Adding sf conf: ${sfaddcmd}"
  echo ${sfaddcmd}  >> ${MLNX_SF_FILE}
}

add_sf_p0(){
  local sf_num=0
  local pci_addr_base="0000:03:00."
  local mac_p0=$(cat /sys/class/net/p0/address)
  if [ -z "${mac_p0}" ]; then return; fi
  base_mac_ecpf0=$(printf '00:04:4b:%02x:%02x\n' $[RANDOM%256] $[RANDOM%256])
  sf_create --device ${pci_addr_base}0 --sfnum $BASE_NET_FACING_SF_NUM --hwaddr $mac_p0
  sf_create --device ${pci_addr_base}0 --sfnum $PF0HPF_SF_ID --hwaddr $base_mac_ecpf0:f0
  # Host facing SFs mapped to ECPF0 VFs
  if ! [ -z "${ECPF0}" ]; then
    for (( i=0; i < $ECPF0; i++))
    do
      mac_idx=$(printf '%02x' $i)
      sf_num=$((ECPF0_VF_SF_BASE_ID + i))
      sf_create --device ${pci_addr_base}0 --sfnum ${sf_num} --hwaddr $base_mac_ecpf0:$mac_idx
    done
  else
    for (( i=0; i < ${#HBN_ARR_REPS_P0[@]}; i++))
    do
      if [ "${HBN_ARR_REPS_P0[${i}]}" = "pf0hpf" ]; then
        continue
      fi
      mac_idx=$(printf '%02x' $i)
      suffix=`echo "${HBN_ARR_REPS_P0[${i}]}" | cut -d"f" -f 3`
      sf_num=$((ECPF0_VF_SF_BASE_ID + suffix))
      sf_create --device ${pci_addr_base}0 --sfnum ${sf_num} --hwaddr $base_mac_ecpf0:$mac_idx
    done
    for (( i=0; i < ${#SFC_ARR_REPS_P0[@]}; i++))
    do
      if [ "${SFC_ARR_REPS_P0[${i}]}" = "pf0hpf" ]; then
        continue
      fi
      mac_idx=$(printf '%02x' $(($i + ${#HBN_ARR_REPS_P0[@]})))
      suffix=`echo "${SFC_ARR_REPS_P0[${i}]}" | cut -d"f" -f 3`
      sf_num=$((ECPF0_VF_SF_BASE_ID + suffix))
      sf_create --device ${pci_addr_base}0 --sfnum ${sf_num} --hwaddr $base_mac_ecpf0:$mac_idx
    done
  fi
}

add_sf_p1(){
  local sf_num=0
  local pci_addr_base="0000:03:00."
  local mac_p1=$(cat /sys/class/net/p1/address)
  if [ -z "${mac_p1}" ]; then return; fi
  base_mac_ecpf1=$(printf '00:04:4b:%02x:%02x\n' $[RANDOM%256] $[RANDOM%256])
  sf_create --device ${pci_addr_base}0 --sfnum $((BASE_NET_FACING_SF_NUM + 1)) --hwaddr $mac_p1
  sf_create --device ${pci_addr_base}0 --sfnum $PF1HPF_SF_ID --hwaddr $base_mac_ecpf1:f1
  # Host facing SFs mapped to ECPF1 VFs
  if ! [ -z "${ECPF1}" ]; then
    for (( i=0; i < $ECPF1; i++))
    do
      mac_idx=$(printf '%02x' $i)
      sf_num=$((ECPF1_VF_SF_BASE_ID + i))
      sf_create --device ${pci_addr_base}0 --sfnum ${sf_num} --hwaddr $base_mac_ecpf1:$mac_idx
    done
  else
    for (( i=0; i < ${#HBN_ARR_REPS_P1[@]}; i++))
    do
      if [ "${HBN_ARR_REPS_P1[${i}]}" = "pf1hpf" ]; then
        continue
      fi
      mac_idx=$(printf '%02x' $i)
      suffix=`echo "${HBN_ARR_REPS_P1[${i}]}" | cut -d"f" -f 3`
      sf_num=$((ECPF1_VF_SF_BASE_ID + suffix))
      sf_create --device ${pci_addr_base}0 --sfnum ${sf_num} --hwaddr $base_mac_ecpf1:$mac_idx
    done
    for (( i=0; i < ${#SFC_ARR_REPS_P1[@]}; i++))
    do
      if [ "${SFC_ARR_REPS_P1[${i}]}" = "pf1hpf" ]; then
        continue
      fi
      mac_idx=$(printf '%02x' $(($i + ${#HBN_ARR_REPS_P1[@]})))
      suffix=`echo "${SFC_ARR_REPS_P1[${i}]}" | cut -d"f" -f 3`
      sf_num=$((ECPF1_VF_SF_BASE_ID + suffix))
      sf_create --device ${pci_addr_base}0 --sfnum ${sf_num} --hwaddr $base_mac_ecpf1:$mac_idx
    done
  fi
}

add_sfs(){
  msg "Clearing existing config in ${MLNX_SF_FILE}";
  > ${MLNX_SF_FILE}

  if ! [ -z "${HBN_COUNT_UPLINKS}" ]; then
    for (( counter=0; counter < ${HBN_COUNT_UPLINKS}; counter++))
    do
      val=${HBN_ARR_UPLINKS[${counter}]}
      if [ "x${val}" == "xp0" ]; then
        add_sf_p0
      elif [ "x${val}" == "xp1" ]; then
        add_sf_p1
      fi
    done
  fi

  if ! [ -z "${SFC_COUNT_UPLINKS}" ]; then
    for (( counter=0; counter < ${SFC_COUNT_UPLINKS}; counter++))
    do
      val=${SFC_ARR_UPLINKS[${counter}]}
      if [ "x${val}" == "xp0" ]; then
        add_sf_p0
      elif [ "x${val}" == "xp1" ]; then
        add_sf_p1
      fi
    done
  fi

  local pci_addr_base="0000:03:00."
  base_mac_dpu=$(printf '00:04:4b:%02x:%02x\n' $[RANDOM%256] $[RANDOM%256])
  for (( i=${BASE_DPU_FACING_SF_NUM}; i < ((${BASE_DPU_FACING_SF_NUM}+${#HBN_ARR_DPU_SFS[@]})); i++))
  do
    mac_idx=$(printf '%02x' $(( ($i & 0xff) )))
    sf_create --device ${pci_addr_base}0 --sfnum ${i} --hwaddr $base_mac_dpu:$mac_idx
  done
  for (( i=((${BASE_DPU_FACING_SF_NUM}+${#HBN_ARR_DPU_SFS[@]})); i < ((${BASE_DPU_FACING_SF_NUM}+${#HBN_ARR_DPU_SFS[@]}+${#SFC_ARR_DPU_SFS[@]})); i++))
  do
    mac_idx=$(printf '%02x' $(( ($i & 0xff) )))
    sf_create --device ${pci_addr_base}0 --sfnum ${i} --hwaddr $base_mac_dpu:$mac_idx
  done
  for (( i=((${BASE_DPU_FACING_SF_NUM}+${#HBN_ARR_DPU_SFS[@]}+${#SFC_ARR_DPU_SFS[@]})); i < ((${BASE_DPU_FACING_SF_NUM}+${#HBN_ARR_DPU_SFS[@]}+${#SFC_ARR_DPU_SFS[@]}+${#SFC_PATCH_PORTS[@]})); i++))
  do
    mac_idx=$(printf '%02x' $(( ($i & 0xff) )))
    sf_create --device ${pci_addr_base}0 --sfnum ${i} --hwaddr $base_mac_dpu:$mac_idx
  done
  msg "==> SF configs saved in $MLNX_SF_FILE"
}

add_dpu_sfs_sf(){
  msg "Adding DPU sfs in the config file ${SFC_CONF_FILE}"
  echo >> ${SFC_CONF_FILE}
  echo 'DPU_SFS_SF=(' >> ${SFC_CONF_FILE}
  for (( i=0; i < ${#HBN_ARR_DPU_SFS[@]}; i++))
  do
    echo "${HBN_ARR_DPU_SFS[${i}]}" >> ${SFC_CONF_FILE}
  done
  for (( i=0; i < ${#SFC_ARR_DPU_SFS[@]}; i++))
  do
    echo "${SFC_ARR_DPU_SFS[${i}]}" >> ${SFC_CONF_FILE}
  done
  for (( i=0; i < ${#SFC_PATCH_PORTS[@]}; i++))
  do
    echo "${SFC_PATCH_PORTS[${i}]}_sf" >> ${SFC_CONF_FILE}
  done
  echo ')' >> ${SFC_CONF_FILE}
  echo '' >> ${SFC_CONF_FILE}
}

add_dpu_sfs_sf_r(){
  msg "Adding DPU sfs reps in the config file ${SFC_CONF_FILE}"
  echo >> ${SFC_CONF_FILE}
  echo 'DPU_SFS_SF_R=(' >> ${SFC_CONF_FILE}
  for (( i=0; i < ${#HBN_ARR_DPU_SFS[@]}; i++))
  do
    echo "${HBN_ARR_DPU_SFS[${i}]}_r" >> ${SFC_CONF_FILE}
  done
  for (( i=0; i < ${#SFC_ARR_DPU_SFS[@]}; i++))
  do
    echo "${SFC_ARR_DPU_SFS[${i}]}_r" >> ${SFC_CONF_FILE}
  done
  for (( i=0; i < ${#SFC_PATCH_PORTS[@]}; i++))
  do
    echo "${SFC_PATCH_PORTS[${i}]}_sf_r" >> ${SFC_CONF_FILE}
  done
  echo ')' >> ${SFC_CONF_FILE}
  echo '' >> ${SFC_CONF_FILE}
}

add_link_propagation(){
  msg "Adding link aggregation reps in the config file ${SFC_CONF_FILE}"
  echo >> ${SFC_CONF_FILE}
  echo 'LINK_PROPAGATION_ARR=(' >> ${SFC_CONF_FILE}
  for (( i=0; i < ${#LINK_PROPAGATION_ARR[@]}; i++))
  do
    echo "${LINK_PROPAGATION_ARR[${i}]}" >> ${SFC_CONF_FILE}
  done
  echo ')' >> ${SFC_CONF_FILE}
  echo '' >> ${SFC_CONF_FILE}
}

remove_auto_ovsbridge_creation(){
    if [ ! -f $MLNX_OVS_FILE ]; then
      die "File: $MLNX_OVS_FILE is missing. Cannot proceed. Aborting!"
    fi
    sed -i 's/CREATE_OVS_BRIDGES=\"yes\"/CREATE_OVS_BRIDGES=\"no\"/g' ${MLNX_OVS_FILE}
    # Try to delete the default bridge and ignore errors
    /usr/bin/ovs-vsctl --timeout=30 --no-wait --if-exists del-br ${BR_HBN_NAME} || true
    /usr/bin/ovs-vsctl --timeout=30 --no-wait --if-exists del-br br-sfc || true
}

update_mnlx_bf_conf(){
    if [ ! -f $MLNX_BF_FILE ]; then
      die "File: $MLNX_BF_FILE is missing. Cannot proceed. Aborting!"
    fi
    # Set ALLOW_SHARED_RQ  to "no"
    sed -i 's/ALLOW_SHARED_RQ=\"yes\"/ALLOW_SHARED_RQ=\"no\"/g' ${MLNX_BF_FILE}
    # Add ESWITCH_MULTIPORT to "yes" if the number of uplinks is 2 or greater
    local sum=$((${HBN_COUNT_UPLINKS} + ${SFC_COUNT_UPLINKS}))

    if [[ "${sum}" -ge "2" ]]; then
      grep -q "ENABLE_ESWITCH_MULTIPORT=" $MLNX_BF_FILE && sed "s/^ENABLE_ESWITCH_MULTIPORT=.*/ENABLE_ESWITCH_MULTIPORT=\"yes\"/" -i $MLNX_BF_FILE ||
      sed "$ a\ENABLE_ESWITCH_MULTIPORT=\"yes\"" -i $MLNX_BF_FILE
    else
      grep -q "ENABLE_ESWITCH_MULTIPORT=" $MLNX_BF_FILE && sed "s/^ENABLE_ESWITCH_MULTIPORT=.*/ENABLE_ESWITCH_MULTIPORT=\"no\"/" -i $MLNX_BF_FILE ||
      sed "$ a\ENABLE_ESWITCH_MULTIPORT=\"no\"" -i $MLNX_BF_FILE
    fi
}

confirm_reboot(){
    read -p "==> Reboot option specified. Please confirm (y/n)? " answer
case ${answer:0:1} in
    y|Y )
        msg " Reboot confirmed. Proceeding..."
        reboot
        exit
    ;;
    * )
        msg "Skipping automatic reboot. Please reboot as early as possible for settings to take effect."
    ;;
esac
}

set_vswitchd_debug_options(){
    msg " Setting debug options for vswitchd..."
    sed -i 's/set ovs_ctl ${1-start} --system-id=random/set ovs_ctl ${1-start} --ovs-vswitchd-options=\"-vdpif_netdev:dbg -vnetdev_offload_dpdk:dbg -vdpdk_offload_doca:dbg -vdpdk:dbg\" --system-id=random/g' /etc/init.d/openvswitch-switch
}

remove_vswitchd_debug_options(){
    msg " Removing debug options for vswitchd..."
    sed -i 's/set ovs_ctl ${1-start} --ovs-vswitchd-options=\"-vdpif_netdev:dbg -vnetdev_offload_dpdk:dbg -vdpdk_offload_doca:dbg -vdpdk:dbg\" --system-id=random/set ovs_ctl ${1-start} --system-id=random/g' /etc/init.d/openvswitch-switch
}

generate_ovs_sf_mapping(){
    tmpfile=$(mktemp)
    cat $SCRIPT_DIR/$OVS_MAPPING_TEMPLATE_FILE > ${tmpfile}

    # the sfc.sh script will look into the generated mapping file
    # and add ports based on order.
    # in case of OVS DPDK/DOCA physical uplinks should be first
    # otherwise it will cause slowness on the OVS side since it
    # has to retry all the invalid ports.
    if ! [ -z "${HBN_COUNT_UPLINKS}" ]; then
      for (( i=0; i < ${HBN_COUNT_UPLINKS}; i++))
        do
            val=${HBN_ARR_UPLINKS[${i}]}
            echo "\"br-hbn~${val}~${val}_sf_r~${val}_sf~${val}_sf_r\"" >> $tmpfile
            LINK_PROPAGATION_ARR+=("${val}:${val}_sf_r")
        done
    fi
    if ! [ -z "${SFC_COUNT_UPLINKS}" ] && ! [ -z "${ENABLE_BR_SFC}" ]; then
      for (( i=0; i < ${SFC_COUNT_UPLINKS}; i++))
        do
            val=${SFC_ARR_UPLINKS[${i}]}
            echo "\"br-sfc~${val}~${val}_sf_r~${val}_sf~${val}_sf_r\"" >> $tmpfile
            LINK_PROPAGATION_ARR+=("${val}:${val}_sf_r")
        done
    fi

    # Generating mappings for SFs to ECPF0 VFs
    if ! [ -z "${ECPF0}" ]; then
      echo "\"br-hbn~pf0hpf~pf0hpf_sf_r~pf0hpf_sf~pf0hpf_sf_r\"" >> $tmpfile
      LINK_PROPAGATION_ARR+=("pf0hpf:pf0hpf_sf_r")
      for (( i=0; i < $ECPF0; i++ ))
      do
          echo "\"br-hbn~pf0vf${i}~pf0vf${i}_sf_r~pf0vf${i}_sf~pf0vf${i}_sf_r\"" >> $tmpfile
          LINK_PROPAGATION_ARR+=("pf0vf${i}:pf0vf${i}_sf_r")
      done
    else
      for (( i=0; i < ${#HBN_ARR_REPS_P0[@]}; i++))
      do
          val=${HBN_ARR_REPS_P0[${i}]}
          echo "\"br-hbn~${val}~${val}_sf_r~${val}_sf~${val}_sf_r\"" >> $tmpfile
          LINK_PROPAGATION_ARR+=("${val}:${val}_sf_r")
      done
      if ! [ -z "${ENABLE_BR_SFC}" ]; then
        for (( i=0; i < ${#SFC_ARR_REPS_P0[@]}; i++))
        do
            val=${SFC_ARR_REPS_P0[${i}]}
            echo "\"br-sfc~${val}~${val}_sf_r~${val}_sf~${val}_sf_r\"" >> $tmpfile
            LINK_PROPAGATION_ARR+=("${val}:${val}_sf_r")
        done
      fi
    fi

    # Generating mappings for SFs to ECPF1 VFs
    if ! [ -z "${ECPF1}" ]; then
      echo "\"br-hbn~pf1hpf~pf1hpf_sf_r~pf1hpf_sf~pf1hpf_sf_r\"" >> $tmpfile
      LINK_PROPAGATION_ARR+=("pf1hpf:pf1hpf_sf_r")
      for (( i=0; i < $ECPF1; i++ ))
      do
          echo "\"br-hbn~pf1vf${i}~pf1vf${i}_sf_r~pf1vf${i}_sf~pf1vf${i}_sf_r\"" >> $tmpfile
          LINK_PROPAGATION_ARR+=("pf1vf${i}:pf1vf${i}_sf_r")
      done
    else
      for (( i=0; i < ${#HBN_ARR_REPS_P1[@]}; i++))
      do
          val=${HBN_ARR_REPS_P1[${i}]}
          echo "\"br-hbn~${val}~${val}_sf_r~${val}_sf~${val}_sf_r\"" >> $tmpfile
          LINK_PROPAGATION_ARR+=("${val}:${val}_sf_r")
      done
      if ! [ -z "${ENABLE_BR_SFC}" ]; then
        for (( i=0; i < ${#SFC_ARR_REPS_P1[@]}; i++))
        do
            val=${SFC_ARR_REPS_P1[${i}]}
            echo "\"br-sfc~${val}~${val}_sf_r~${val}_sf~${val}_sf_r\"" >> $tmpfile
            LINK_PROPAGATION_ARR+=("${val}:${val}_sf_r")
        done
      fi
    fi

    for (( i=0; i < ${#HBN_ARR_DPU_SFS[@]}; i+=2))
    do
        val=${HBN_ARR_DPU_SFS[${i}]}
        temp=$((i + 1))
        val1=${HBN_ARR_DPU_SFS[${temp}]}
        echo '"''br-hbn~'${val}'_r~'${val1}'_r~'${val1}'~'${val1}'_r"' >> $tmpfile
        LINK_PROPAGATION_ARR+=("${val}_r:${val1}_r")
    done
    if ! [ -z "${ENABLE_BR_SFC}" ]; then
      for (( i=0; i < ${#SFC_ARR_DPU_SFS[@]}; i+=2))
      do
          val=${SFC_ARR_DPU_SFS[${i}]}
          temp=$((i + 1))
          val1=${SFC_ARR_DPU_SFS[${temp}]}
          echo '"''br-sfc~'${val}'_r~'${val1}'_r~'${val1}'~'${val1}'_r"' >> $tmpfile
          LINK_PROPAGATION_ARR+=("${val}_r:${val1}_r")
      done
    fi
    if ! [ -z "${ENABLE_BR_SFC}" ]; then
      for (( i=0; i < ${#SFC_PATCH_PORTS[@]}; i++))
      do
          val=${SFC_PATCH_PORTS[${i}]}
          echo '"''Patch~'${val}'_sf_r~'${val}'_sf_r~'${val}'_sf~'${val}'_sf_r"' >> $tmpfile
      done
    fi

    # Add closing paranthesis.
    echo ")" >> $tmpfile

    # Define sfc bridge if enabled.
    if ! [ -z "${ENABLE_BR_SFC}" ]; then
      echo "BR_SFC_NAME=br-sfc" >> $tmpfile
    fi

    if ! [ -z "${ENABLE_BR_SFC_DEFAULT_FLOWS}" ]; then
      echo "ENABLE_BR_SFC_DEFAULT_FLOWS=yes" >> $tmpfile
    fi

    msg "==> Generated Mapping file"
    cat $tmpfile

    cp $tmpfile $SFC_CONF_FILE
    rm -rf $tmpfile

}

generate_hbn_configuration(){
  msg "==> Generating hbn configuration..."
  HBN_ARR_DPU_SFS_ORIG="${HBN_ARR_DPU_SFS_ORIG[@]}" HBN_ARR_UPLINKS="${HBN_ARR_UPLINKS[@]}" \
  HBN_ARR_REPS_P0="${HBN_ARR_REPS_P0[@]}" HBN_ARR_REPS_P1="${HBN_ARR_REPS_P1[@]}" \
  SFC_ARR_DPU_SFS_ORIG="${SFC_ARR_DPU_SFS_ORIG[@]}" SFC_ARR_UPLINKS="${SFC_ARR_UPLINKS[@]}" \
  SFC_ARR_REPS_P0="${SFC_ARR_REPS_P0[@]}" SFC_ARR_REPS_P1="${SFC_ARR_REPS_P1[@]}" \
  SFC_PATCH_PORTS="${SFC_PATCH_PORTS[@]}" LINK_PROPAGATION_ARR="${LINK_PROPAGATION_ARR[@]}" \
  ENABLE_BR_SFC="${ENABLE_BR_SFC}" ENABLE_BR_SFC_DEFAULT_FLOWS="${ENABLE_BR_SFC_DEFAULT_FLOWS}"\
  python3 $SCRIPT_DIR/generate_hbn_conf.py
}

install_sfc_service(){
    msg "==> Installing sfc systemd service..."
    cp $SCRIPT_DIR/sfc.service /etc/systemd/system/sfc.service
    msg "==> Finished installing sfc systemd service."
    msg "==> Enabling sfc systemd service"
    systemctl daemon-reload
    systemctl enable sfc
    systemctl enable sfc-state-propagation.service
}

use_hbn_config_file(){
  msg "==> Fetching input from configuration file"
  python3 $SCRIPT_DIR/use_hbn_conf.py
  BR_HBN_UPLINKS=`cat /tmp/.BR_HBN_UPLINKS`
  BR_HBN_REPS=`cat /tmp/.BR_HBN_REPS`
  BR_HBN_SFS=`cat /tmp/.BR_HBN_SFS`
  BR_SFC_UPLINKS=`cat /tmp/.BR_SFC_UPLINKS`
  BR_SFC_REPS=`cat /tmp/.BR_SFC_REPS`
  BR_SFC_SFS=`cat /tmp/.BR_SFC_SFS`
  BR_HBN_SFC_PATCH_PORTS=`cat /tmp/.BR_HBN_SFC_PATCH_PORTS`
  LINK_PROPAGATION=`cat /tmp/.LINK_PROPAGATION`
  ENABLE_BR_SFC=`cat /tmp/.ENABLE_BR_SFC`
  ENABLE_BR_SFC_DEFAULT_FLOWS=`cat /tmp/.ENABLE_BR_SFC_DEFAULT_FLOWS`
}

copy_files(){
    msg "==> Start copying udev SF rename scripts"
    # Backup original files first
    cp $UDEV_SCRIPT_PATH/$SF_RENAME_SCRIPT{,.original}
    cp $SCRIPT_DIR/$SF_RENAME_SCRIPT $UDEV_SCRIPT_PATH/
    cp $UDEV_SCRIPT_PATH/$SF_REP_RENAME_SCRIPT{,.original}
    cp $SCRIPT_DIR/$SF_REP_RENAME_SCRIPT $UDEV_SCRIPT_PATH/
    msg "==> Finished copying udev SF rename scripts"

    msg "==> Start copying sysctl conf"
    mkdir -p $SYSCTL_PATH
    cp $SCRIPT_DIR/$SYSCTL_CONF $SYSCTL_PATH/
    msg "==> Finished copying sysctl conf file"

    msg "==> Start copying ovs sfc files"
    # We do not need to copy sfc.conf since it is done by the generate_ovs_sf_mapping function.
    mkdir -p /opt/mellanox/sfc
    cp $SCRIPT_DIR/sfc.sh  /opt/mellanox/sfc/
    chmod +x /opt/mellanox/sfc/sfc.sh
    msg "==> Finished copying ovs sfc files"

    cp $SCRIPT_DIR/ovs-vswitchd-watchdog.py  /opt/mellanox/sfc/
    msg "==> Finished copying ovs-vswitchd-watchdog.py"
}

install_sfc_hbn_software() {
	msg "==> Install SFC HBN packages"

  # The below change was added to force packages to install from local repo
  # instead of online repo.
  # See https://gitlab-master.nvidia.com/hareeshp/sfc-hbn/-/merge_requests/16

  # apt install -y hbn-runtime
  if [ "X$distro" = "Xubuntu2004" ]; then
    DEBIAN_FRONTEND=noninteractive apt install -y --allow-downgrades /var/hbn-repo-aarch64-ubuntu2004-local/*.deb
  fi

  if [ "X$distro" = "Xubuntu2204" ]; then
    DEBIAN_FRONTEND=noninteractive apt install -y --allow-downgrades /var/hbn-repo-aarch64-ubuntu2204-local/*.deb
  fi

	msg "==> Finished install SFC HBN packages"
}

configure_mgmt_vrf() {
    msg "==> Start copying ifup hooks"
    # IFUP_HOOK_50 used to create mgmt VRF
    # mgmt VRF is supported by NetworkManager in Ubuntu22.04
    mkdir -p $IFUP_HOOKS_PATH
    cp $SCRIPT_DIR/$IFUP_HOOK_50 $IFUP_HOOKS_PATH/
    chmod +x $IFUP_HOOKS_PATH/$IFUP_HOOK_50
    msg "==> Finished copying ifup hooks files"
    msg "==> Start mgmt VRF configuration"
    # Remove cloud-init configuration to prevent netplan files creation
    rm -f /var/lib/cloud/seed/nocloud-net/network-config
    if [ -e /etc/netplan/50-cloud-init.yaml ]; then
      rm -f /etc/netplan/50-cloud-init.yaml
      netplan apply > /dev/null 2>&1 || true
    fi
    # Change oob_net0 configuration using networkd
    cat >> /etc/systemd/network/25-oob-net0.network << EOF
[Match]
Name=oob_net0

[Network]
VRF=mgmt
DHCP=ipv4

[DHCP]
UseDomains=true
EOF
    # Set tmfifo_net0 configuration using networkd
    cat >> /etc/systemd/network/25-tmfifo-net0.network << EOF
[Match]
Name=tmfifo_net0

[Network]
VRF=mgmt
Address=192.168.100.2/30

[Route]
Gateway=192.168.100.1
Metric=1025

[DHCP]
UseDomains=true
EOF
    # Create management VRF
    cat >> /etc/systemd/network/25-mgmt.network << EOF
[Match]
Name=mgmt

[Network]
Address=127.0.0.1/8
Address=::1/128
EOF

  cat >> /etc/systemd/network/25-mgmt.netdev << EOF
[NetDev]
Name=mgmt
Kind=vrf

[VRF]
TableId=1001
EOF
    # Disable containerd and kubelet services
    systemctl disable kubelet
    systemctl disable containerd
    systemctl disable docker
    systemctl disable ssh
    systemctl disable ntp
    # Disable NetworkManager services
    systemctl disable NetworkManager.service
    systemctl disable NetworkManager-wait-online.service
    # Enable containerd and kubelet services for management VRF
    # The actual service files created by /lib/systemd/system-generators/systemd-vrf-generator
    mkdir -p /etc/systemd/system/vrf@mgmt.target.wants
    ln -snf /etc/systemd/system/containerd@.service /etc/systemd/system/vrf@mgmt.target.wants/containerd@mgmt.service
    ln -snf /etc/systemd/system/kubelet@.service /etc/systemd/system/vrf@mgmt.target.wants/kubelet@mgmt.service
    ln -snf /etc/systemd/system/docker@.service /etc/systemd/system/vrf@mgmt.target.wants/docker@mgmt.service
    ln -snf /etc/systemd/system/ntp@.service /etc/systemd/system/vrf@mgmt.target.wants/ntp@mgmt.service
    sed -i -e 's@ -H fd://@@' -e 's/Wants=containerd.service/Wants=containerd@mgmt.service/' /lib/systemd/system/docker.service
    # Generate services files for management VRF
    /lib/systemd/system-generators/systemd-vrf-generator /run/systemd/generator /run/systemd/generator.early /run/systemd/generator.late
    # Start containerd, docker and kubelet services onthe first boot after the deployment
    cat > $SCRIPT_DIR/services_helper << EOF
#!/bin/bash

systemctl daemon-reload
systemctl enable ssh@mgmt.service
systemctl start ssh@mgmt.service
systemctl start ntp@mgmt.service
systemctl start containerd@mgmt.service
systemctl start kubelet@mgmt.service
systemctl start docker@mgmt.service
EOF
    chmod +x $SCRIPT_DIR/services_helper
    sed -i -e "/runcmd:/a\ \ - [ $SCRIPT_DIR/services_helper ]" /var/lib/cloud/seed/nocloud-net/user-data
    sed -i "s|ExecStart=|ExecStartPre=$SCRIPT_DIR/services_helper\nExecStart=|g" /lib/systemd/system/cloud-config.service
    # Prevent from import_doca_telemetry.sh enabling kubelet and containerd services
    perl -ni -e 'print unless /systemctl/' /opt/mellanox/doca/services/telemetry/import_doca_telemetry.sh
    # Use systemd-resolved to keep /etc/resolv.conf up-to-date
    rm -f /etc/resolv.conf
    ln -snf /run/systemd/resolve/resolv.conf /etc/resolv.conf
    # Create .bash_aliases to include VRF in the shell prompt
    cat >> /root/.bash_aliases << 'EOF'
NS=$(ip netns identify)
[ -n "$NS" ] && NS=":${NS}"

VRF=$(ip vrf identify)
[ -n "$VRF" ] && VRF=":${VRF}"

PS1='${debian_chroot:+($debian_chroot)}\u@\h${NS}${VRF}:\w\$ '
EOF
    msg "==> Finished mgmt VRF configuration"
}

disable_netplan() {
  msg "==> Disable netplan configuration by cloud-init"
  echo "network: {config: disabled}" > /etc/cloud/cloud.cfg.d/99-custom-networking.cfg
  /bin/rm -f /etc/netplan/*
}

grub_cmd_update() {
  if ! grep -wq "cgroup_no_v1=net_prio,net_cls," /etc/default/grub; then
    sed -i -r -e 's/^(GRUB_CMDLINE_LINUX=.*)\"/\1 cgroup_no_v1=net_prio,net_cls,"/' /etc/default/grub
  fi

  if grep -Pwq "hugepagesz=.*[0-9]kB" /etc/default/grub; then
    sed -i -r -e 's/hugepagesz=.*[0-9]kB/hugepagesz='"${HUGEPAGESZ}"'kB/g' /etc/default/grub
  else
    sed -i -r -e 's/^(GRUB_CMDLINE_LINUX=.*)\"/\1 hugepagesz='"${HUGEPAGESZ}"'kB"/' /etc/default/grub
  fi

  if grep -Pwq "hugepages=.*[0-9]" /etc/default/grub; then
    sed -i -r -e 's/hugepages=.*[0-9]/hugepages='"${HUGEPAGES}"'/g' /etc/default/grub
  else
    sed -i -r -e 's/^(GRUB_CMDLINE_LINUX=.*)\"/\1 hugepages='"${HUGEPAGES}"'"/' /etc/default/grub
  fi

  update-grub > /dev/null 2>&1
}

cx_nvconfig(){
    msg "==> Start CX device configuration"
    RC=0

    for dev in `lspci -nD -d 15b3: | grep 'a2d[26c]' | cut -d ' ' -f 1`
    do
        msg "==> Running mlxconfig for $dev"
        output=`mlxconfig -y -d $dev set \
            PF_BAR2_ENABLE=$PF_BAR2_ENABLE \
            PER_PF_NUM_SF=$PER_PF_NUM_SF \
            PF_TOTAL_SF=$PF_TOTAL_SF \
            PF_SF_BAR_SIZE=$PF_SF_BAR_SIZE \
            NUM_PF_MSIX_VALID=$NUM_PF_MSIX_VALID \
            PF_NUM_PF_MSIX_VALID=$PF_NUM_PF_MSIX_VALID \
            PF_NUM_PF_MSIX=$PF_NUM_PF_MSIX \
            LAG_RESOURCE_ALLOCATION=1; \
            echo Status: $?`
        rc=${output:0-1}
        RC=$((RC+rc))
        msg "==> MLXCONFIG:"
        msg "$output"
        if [ $rc -ne 0 ]; then
            msg "==> Failed to configure CX device: $dev"
        fi
    done

    msg "==> Finished CX device configuration. Status: $RC"
    msg "==> mlxfwreset or power cycle is required"
    return $RC
}

add_cni_config(){

  # Copying CNI config
  msg "==> Copying CNI configuration"
  cp $SCRIPT_DIR/cni/$CNI_CONF  $CNI_CONF_PATH

}

detect_uplinks(){
  # Set global configuration variables after validating the uplinks
  if ! [ -z "${BR_HBN_UPLINKS-x}" ]; then
    BR_HBN_UPLINKS=${BR_HBN_UPLINKS:-"p0,p1"}
  fi
  HBN_ARR_UPLINKS=(${BR_HBN_UPLINKS//,/ })
  BR_SFC_UPLINKS=${BR_SFC_UPLINKS:-""}
  SFC_ARR_UPLINKS=(${BR_SFC_UPLINKS//,/ })
  HBN_COUNT_UPLINKS=${#HBN_ARR_UPLINKS[@]}
  SFC_COUNT_UPLINKS=${#SFC_ARR_UPLINKS[@]}
  SFC_COUNT_UPLINKS=${SFC_COUNT_UPLINKS:-0}
  if ! [ -z "${BR_HBN_REPS-x}" ]; then
    BR_HBN_REPS=${BR_HBN_REPS:-"pf0hpf,pf1hpf,pf0vf0-pf0vf13"}
  fi
  BR_SFC_REPS=${BR_SFC_REPS:-""}
  HBN_ARR_REPS_P0=()
  HBN_ARR_REPS_P1=()
  SFC_ARR_REPS_P0=()
  SFC_ARR_REPS_P1=()
  if ! [ -z "${BR_HBN_SFS-x}" ]; then
    BR_HBN_SFS=${BR_HBN_SFS:-"pf0dpu1,pf0dpu3"}
  fi
  BR_SFC_SFS=${BR_SFC_SFS:-""}
  HBN_ARR_DPU_SFS=()
  HBN_ARR_DPU_SFS_ORIG=()
  SFC_ARR_DPU_SFS=()
  SFC_ARR_DPU_SFS_ORIG=()
  BR_HBN_SFC_PATCH_PORTS=${BR_HBN_SFC_PATCH_PORTS:-""}
  SFC_PATCH_PORTS=()
  LINK_PROPAGATION=${LINK_PROPAGATION:-""}
  LINK_PROPAGATION_ARR=()
  HUGEPAGESZ=${HUGEPAGE_SIZE:-2048}
  HUGEPAGES=${HUGEPAGE_COUNT:-3072}
}

validate_hbn_configuration(){
  python3 $SCRIPT_DIR/validate_hbn_conf.py
}

override_hbn_reps(){
  # Generating mappings for SFs to ECPF1 VFs
  if ! [ -z "${ECPF0}" ] || ! [ -z "${ECPF1}" ]; then
    BR_HBN_REPS=""
  fi
  if ! [ -z "${ECPF0}" ] && [ ${ECPF0} -gt 0 ]; then
    vf_p0_num=$(($ECPF0-1))
    BR_HBN_REPS="${BR_HBN_REPS}pf0vf0-pf0vf${vf_p0_num},"
  fi
  if ! [ -z "${ECPF1}" ] && [ ${ECPF1} -gt 0 ]; then
    vf_p1_num=$(($ECPF1-1))
    BR_HBN_REPS="${BR_HBN_REPS}pf1vf0-pf1vf${vf_p1_num},"
  fi
}

expand_hbn_sf_vf_variables(){
  # Expand and group VFs
  override_hbn_reps
  local arr_reps_temp=(${BR_HBN_REPS//,/ })
  local count_arr_reps_temp=${#arr_reps_temp[@]}
  for (( i=0; i< ${count_arr_reps_temp}; i++ )); do
    local val=${arr_reps_temp[${i}]}
    local chk_prefix=`echo "${val}" | cut -d"v" -f 1`
    if [ "${chk_prefix}" = "pf0hpf" ]; then
      HBN_ARR_REPS_P0+=("pf0hpf")
      continue
    fi
    if [ "${chk_prefix}" = "pf1hpf" ]; then
      HBN_ARR_REPS_P1+=("pf1hpf")
      continue
    fi
    if  [[ ${arr_reps_temp[${i}]} =~ "-" ]]; then
      local prefix=`echo "${val}" | cut -d"-" -f 1 | cut -d"f" -f 1`
      prefix+="f"
      prefix+=`echo "${val}" | cut -d"-" -f 1 | cut -d"f" -f 2`
      prefix+="f"
      min=`echo "$val" | cut -d"-" -f 1 | cut -d"f" -f 3`
      max=`echo "$val" | cut -d"-" -f 2 | cut -d"f" -f 3`
      for (( j=${min}; j<=${max}; j++)); do
        if [ "${chk_prefix}" = "pf0" ]; then
          HBN_ARR_REPS_P0+=(${prefix}${j})
        elif [ "${chk_prefix}" = "pf1" ]; then
          HBN_ARR_REPS_P1+=(${prefix}${j})
        fi
      done
    else
      if [ "${chk_prefix}" = "pf0" ]; then
        HBN_ARR_REPS_P0+=(${val})
      elif [ "${chk_prefix}" = "pf1" ]; then
        HBN_ARR_REPS_P1+=(${val})
      fi
    fi
  done
  # Expand and group SFs
  local arr_dpu_sfs_temp=(${BR_HBN_SFS//,/ })
  for (( i=0; i< ${#arr_dpu_sfs_temp[@]}; i++ )); do
    local val=${arr_dpu_sfs_temp[${i}]}
    local suffix=`echo "${val}" | cut -d"u" -f 2`
    local prefix=`echo "${val}" | cut -d"u" -f 1`
    prefix+="u"
    HBN_ARR_DPU_SFS+=(${prefix}$((suffix - 1))_sf)
    HBN_ARR_DPU_SFS+=(${val}_sf)
    HBN_ARR_DPU_SFS_ORIG+=(${val}_sf)
  done
  # Expand link aggregation
  local arr_link_aggregation=(${LINK_PROPAGATION//,/ })
  for (( i=0; i< ${#arr_link_aggregation[@]}; i++ )); do
    local val=${arr_link_aggregation[${i}]}
    LINK_PROPAGATION_ARR+=(${val})
  done
}

expand_sfc_sf_vf_variables(){
  # Expand and group VFs
  local arr_reps_temp=(${BR_SFC_REPS//,/ })
  local count_arr_reps_temp=${#arr_reps_temp[@]}
  for (( i=0; i< ${count_arr_reps_temp}; i++ )); do
    local val=${arr_reps_temp[${i}]}
    local chk_prefix=`echo "${val}" | cut -d"v" -f 1`
    if [ "${chk_prefix}" = "pf0hpf" ]; then
      SFC_ARR_REPS_P0+=("pf0hpf")
      continue
    fi
    if [ "${chk_prefix}" = "pf1hpf" ]; then
      SFC_ARR_REPS_P1+=("pf1hpf")
      continue
    fi
    if  [[ ${arr_reps_temp[${i}]} =~ "-" ]]; then
      local prefix=`echo "${val}" | cut -d"-" -f 1 | cut -d"f" -f 1`
      prefix+="f"
      prefix+=`echo "${val}" | cut -d"-" -f 1 | cut -d"f" -f 2`
      prefix+="f"
      min=`echo "$val" | cut -d"-" -f 1 | cut -d"f" -f 3`
      max=`echo "$val" | cut -d"-" -f 2 | cut -d"f" -f 3`
      for (( j=${min}; j<=${max}; j++)); do
        if [ "${chk_prefix}" = "pf0" ]; then
          SFC_ARR_REPS_P0+=(${prefix}${j})
        elif [ "${chk_prefix}" = "pf1" ]; then
          SFC_ARR_REPS_P1+=(${prefix}${j})
        fi
      done
    else
      if [ "${chk_prefix}" = "pf0" ]; then
        SFC_ARR_REPS_P0+=(${val})
      elif [ "${chk_prefix}" = "pf1" ]; then
        SFC_ARR_REPS_P1+=(${val})
      fi
    fi
  done
  # Expand and group SFs
  local arr_dpu_sfs_temp=(${BR_SFC_SFS//,/ })
  for (( i=0; i< ${#arr_dpu_sfs_temp[@]}; i++ )); do
    local val=${arr_dpu_sfs_temp[${i}]}
    local suffix=`echo "${val}" | cut -d"u" -f 2`
    local prefix=`echo "${val}" | cut -d"u" -f 1`
    prefix+="u"
    SFC_ARR_DPU_SFS+=(${prefix}$((suffix - 1))_sf)
    SFC_ARR_DPU_SFS+=(${val}_sf)
    SFC_ARR_DPU_SFS_ORIG+=(${val}_sf)
  done
  # Expand and group custom interfaces
  local arr_patch_temp=(${BR_HBN_SFC_PATCH_PORTS//,/ })
  for (( i=0; i< ${#arr_patch_temp[@]}; i++ )); do
    SFC_PATCH_PORTS+=(${arr_patch_temp[${i}]})
  done
}

###########################################
###             Main                    ###
###########################################
ARGS="$@"
msg ""
msg "#####################################################################################"
msg "==> Starting OVS SFC Installation"
msg "==> Script called as:$0 $ARGS"

install_sfc_hbn_software

if [ "${CONFIG_OPTIONS}" = "true" ]; then
  use_hbn_config_file
fi

detect_uplinks
validate_hbn_configuration

expand_hbn_sf_vf_variables
expand_sfc_sf_vf_variables
remove_auto_ovsbridge_creation
update_mnlx_bf_conf

add_sfs
generate_ovs_sf_mapping
generate_hbn_configuration

copy_files

add_cni_config
install_sfc_service
if [[ -z "$CLOUD_OPTION" ]]; then
  if [[ "X$distro" =~ Xubuntu2[02]04 ]]; then
    # Configure management VRF for Ubuntu 20.04 and 22.04
    configure_mgmt_vrf
    disable_netplan
  fi
fi
grub_cmd_update
cx_nvconfig

if [ "${DEBUG_OPTION}" = "true" ]; then
  set_vswitchd_debug_options
else
  remove_vswitchd_debug_options
fi

if [ "${REBOOT_OPTION}" = "true" ]; then
  confirm_reboot
fi

add_dpu_sfs_sf
add_dpu_sfs_sf_r
add_link_propagation

msg "Success. OVS SFC Installation completed"
msg ""
msg "#####################################################################################"
msg "                            NOTICE                                                   "
msg "OVS SFC install completed. Please reboot at the earliest for settings to take effect."
msg "#####################################################################################"

