#!/usr/bin/env bash

###############################################################################
# Copyright (c) 2022 NVIDIA Corporation. All rights reserved.
#
# This Software is licensed under one of the following licenses:
#
# 1) under the terms of the "Common Public License 1.0" a copy of which is
#    available from the Open Source Initiative, see
#    http://www.opensource.org/licenses/cpl.php.
#
# 2) under the terms of the "The BSD License" a copy of which is
#    available from the Open Source Initiative, see
#    http://www.opensource.org/licenses/bsd-license.php.
#
# 3) under the terms of the "GNU General Public License (GPL) Version 2" a
#    copy of which is available from the Open Source Initiative, see
#    http://www.opensource.org/licenses/gpl-license.php.
#
# Licensee has the right to choose one of the above licenses.
#
# Redistributions of source code must retain the above copyright
# notice and one of the license notices.
#
# Redistributions in binary form must reproduce both the above copyright
# notice, one of the license notices in the documentation
# and/or other materials provided with the distribution.
###############################################################################

# We do not set safe bash options intentionally since some ovs commands do return non-zero exit code.
# set -euo pipefail

die() {
  local msg=$1
  local code=${2-1} # default exit status 1
  msg "Error: $msg"
  exit "$code"
}

msg() {
  echo >&2 -e "${1-}"
}

function retry {
  local retries=$1
  shift

  local count=0
  until "$@"; do
    exit=$?
    wait=$((2 ** $count))
    count=$(($count + 1))
    if [ $count -lt $retries ]; then
      msg "Retry $count/$retries exited $exit, retrying in $wait seconds..."
      sleep $wait
    else
      msg "Retry $count/$retries exited $exit, no more retries left."
      return $exit
    fi
  done
  return 0
}


###  OVS functions
OVS_TIMEOUT=60   # seconds to timeout ovs operation, doca port addition is sometime slow changing from 30 to 60
OFCTL_TIMEOUT=60  # Openflow timeout, doca operations sometime block the main thread for 25 seconds increasing the number from 30 to 60
VSCTL="/usr/bin/ovs-vsctl --timeout=${OVS_TIMEOUT}"
OFCTL="/usr/bin/ovs-ofctl --timeout=${OFCTL_TIMEOUT}"
BR_HBN_NAME=${BR_HBN_NAME:-br-hbn}
BR_HBN_NAME=${BR_SFC_NAME:""}
HBN_OF_PORT_INDEX=1
SFC_OF_PORT_INDEX=1
MTU_UPLINK=${MTU_UPLINK:-9702}
MTU_INTERFACES=${MTU_INTERFACES:-9216}

check_br(){
    ${VSCTL} br-exists $1
    retval=$?
    if [ ${retval} -ne 0 ]; then
        msg "Bridge $1 not present"
        return ${retval}
    fi
}

create_bridge() {
    local bridgename=$1
    ${VSCTL} --may-exist add-br ${bridgename} -- set bridge ${bridgename} datapath_type=netdev -- set-fail-mode ${bridgename} secure
    local index=$(${VSCTL} list interface vxlan0 2>/dev/null | grep ofport_request | cut -d ":" -f2 | xargs)
    if ! [ -z "${index}" ]; then
        msg "Port vxlan0 is already added to bridge ${bridgename} and configured with ofport_request=${index}"
        return
    fi
    if [ "x${bridgename}" == "x${BR_HBN_NAME}" ]; then
      ${VSCTL} --may-exist add-port ${bridgename} vxlan0 -- set interface vxlan0 type=vxlan ofport_request=${HBN_OF_PORT_INDEX} \
        options:explicit=true options:tos=inherit options:remote_ip=flow -- set bridge ${bridgename} \
        external_ids:ofport_index=$((${HBN_OF_PORT_INDEX}+1))
        HBN_OF_PORT_INDEX=$((${HBN_OF_PORT_INDEX}+1))
    fi
}

ensure_bridge() {
  local bridgename=$1
  msg "Ensuring bridge ${bridgename} is present"
  create_bridge ${bridgename}
  retry 5 check_br ${bridgename}
  retval=$?
    if [ ${retval} -ne 0 ]; then
        msg "Bridge $1 is not present. See Openvswitch logs. Exiting"
        exit -1
    fi
}

delete_bridge() {
    local bridgename=$1
    ${VSCTL} --if-exists del-br ${bridgename}
}

del_flows(){
    local bridge=${1:-br-hbn}
    ${OFCTL} del-flows ${bridge}
}

set_ofport_index(){
  local bridgename=$1
  local -n of_port_index=${2}
  local index=$(${VSCTL} --data=bare list bridge ${bridgename} | grep ofport_index | cut -d "=" -f2)
  if ! [ -z "${index}" ]; then
    msg "ofport_index on bridge ${bridgename} is :${index}"
    of_port_index=${index}
  fi
}

initialize(){
    delete_bridge ovsbr1
    delete_bridge ovsbr2
    local restart=""
    msg "==> Ensuring OVS Hardware offload is enabled"
    ovs_hw_offload=$(${VSCTL} get Open_vSwitch . other_config:hw-offload 2> /dev/null | tr -d '\"')
    if [ "X${ovs_hw_offload}" != "Xtrue" ]; then
        ${VSCTL} set Open_vSwitch . other_config:hw-offload=true
        msg "Marking OVS for restart after setting hw-offload to true "
        restart="yes"
    fi
    msg "==> Completed OVS Hardware offload configuration"
    msg "==> Ensuring DPDK is enabled"
    ovs_dpdk=$(${VSCTL} get Open_vSwitch . other_config:dpdk-init 2> /dev/null | tr -d '\"')
    if [ "X${ovs_dpdk}" != "Xtrue" ]; then
        ${VSCTL} set Open_vSwitch . other_config:dpdk-init=true
        msg "Marking OVS for restart after setting dpdk-init to true "
        restart="yes"
    fi
    msg "==> Completed DPDK configuration"
    msg "==> Ensuring DOCA is enabled"
    ovs_doca=$(${VSCTL} get Open_vSwitch . other_config:doca-init 2> /dev/null | tr -d '\"')
    if [ "X${ovs_doca}" != "Xtrue" ]; then
        ${VSCTL} set Open_vSwitch . other_config:doca-init=true
        msg "Marking OVS for restart after setting doca-init to true "
        restart="yes"
    fi
    msg "==> Completed DOCA configuration"
    msg "==> Ensuring DPDK-EXTRA is configured"
    ovs_doca_extra=$(${VSCTL} get Open_vSwitch . other_config:dpdk-extra 2> /dev/null | tr -d '\"')
    if [ "X${ovs_doca_extra}" != "X-a 0000:00:00.0" ]; then
        ${VSCTL} set Open_vSwitch . other_config:dpdk-extra="-a 0000:00:00.0"
        msg "Marking OVS for restart after setting dpdk-extra to \"-a 0000:00:00.0\" "
        restart="yes"
    fi
    msg "==> Completed DPDK-EXTRA configuration"
    msg "==> Ensuring CTL PIPE size is configured"
    ovs_ctl_pipe_size=$(${VSCTL} get Open_vSwitch . other_config:ctl-pipe-size 2> /dev/null | tr -d '\"')
    if [ "X${ovs_ctl_pipe_size}" != "X64" ]; then
        ${VSCTL} set Open_vSwitch . other_config:ctl-pipe-size=64
    fi
    msg "==> Completed CTL PIPE size configuration"
    msg "==> Ensuring HW OFFLOAD CT size is configured"
    ovs_hw_offload_ct_size=$(${VSCTL} get Open_vSwitch . other_config:hw-offload-ct-size 2> /dev/null | tr -d '\"')
    if [ "X${ovs_hw_offload_ct_size}" != "X0" ]; then
        ${VSCTL} set Open_vSwitch . other_config:hw-offload-ct-size=0
        msg "Marking OVS for restart after setting hw-offload-ct-size to 0"
        restart="yes"
    fi
    msg "==> Completed HW OFFLOAD CT size configuration"
    msg "==> Ensuring PMD-quiet-idle is configured"
    ovs_pmd_quiet_idle=$(${VSCTL} get Open_vSwitch . other_config:pmd-quiet-idle 2> /dev/null | tr -d '\"')
    if [ "X${ovs_pmd_quiet_idle}" != "Xtrue" ]; then
        ${VSCTL} set Open_vSwitch . other_config:pmd-quiet-idle=true
        msg "Marking OVS for restart after setting pmd-quiet-idle to true "
        restart="yes"
    fi
    msg "==> Completed PMD-quiet-idle size configuration"
    msg "==> Ensuring DPDK max memory zone is configured"
    ovs_dpdk_memzone=$(${VSCTL} get Open_vSwitch . other_config:dpdk-max-memzones 2> /dev/null | tr -d '\"')
    if [ "X${ovs_dpdk_memzone}" != "X50000" ]; then
        ${VSCTL} set Open_vSwitch . other_config:dpdk-max-memzones=50000
        msg "Marking OVS for restart after setting DPDK max memory zone"
        restart="yes"
    fi
    msg "==> Completed DPDK max memory zone configuration"
    if [ "X${restart}" == "Xyes" ]; then
        systemctl restart openvswitch-switch
        msg "Exiting after setting the required other_config"
        exit 1 # Bail out and wait for the script to be called again
    fi
    ${VSCTL} set Open_vSwitch . other_config:max-idle=60000
    ${VSCTL} set Open_vSwitch . other_config:max-revalidator=2000

    set_ofport_index ${BR_HBN_NAME} HBN_OF_PORT_INDEX
    ensure_bridge ${BR_HBN_NAME}
    if ! [ -z "${BR_SFC_NAME}" ]; then
      set_ofport_index ${BR_SFC_NAME} SFC_OF_PORT_INDEX
      ensure_bridge ${BR_SFC_NAME}
    fi

    # clear flows previously created
    ${OFCTL} del-flows ${BR_HBN_NAME} "table=0,cookie=0/-1"
}

# Add_port() adds given port to the given bridge and sets openflow `ofport_request`
# index based on of_port_index
add_port(){
    local port=$1
    local bridge=${2:-br-hbn}
    local -n of_port_index=${3}

    local rep_index=$(cat /sys/class/net/${port}/phys_port_name | grep pf | cut -d "f" -f 3)
    local phys_port_name=$(cat /sys/class/net/${port}/phys_port_name)
    local index=$(${VSCTL} list interface $port 2>/dev/null | grep ofport_request | cut -d ":" -f2 | xargs)
    if ! [ -z "${index}" ]; then
        msg "Port is already added to bridge ${bridge} and configured with ofport_request=${index}"
        return
    fi
    msg "Adding port: ${port} to bridge: ${bridge} with index ${of_port_index}"
    if [ -z "${rep_index}" ]; then
        if [ "X${phys_port_name}" == "Xp1" ]; then
            ${VSCTL} --may-exist add-port ${bridge} ${port}  -- set interface ${port} type=dpdk ofport_request=${of_port_index} \
              -- set interface ${port} mtu_request=${MTU_UPLINK} -- set bridge ${bridge} external_ids:ofport_index=$((${of_port_index}+1))
        elif [ "X${phys_port_name}" == "Xp0" ]; then
            ${VSCTL} --may-exist add-port ${bridge} ${port}  -- set interface ${port} type=dpdk ofport_request=${of_port_index} \
              -- set interface ${port} mtu_request=${MTU_UPLINK} -- set bridge ${bridge} external_ids:ofport_index=$((${of_port_index}+1))
        else
            ${VSCTL} --may-exist add-port ${bridge} ${port}  -- set interface ${port} type=dpdk ofport_request=${of_port_index} \
              -- set interface ${port} mtu_request=${MTU_INTERFACES} -- set bridge ${bridge} external_ids:ofport_index=$((${of_port_index}+1))
        fi
    else
        rep_index=$(cat /sys/class/net/${port}/phys_port_name | grep vf)
        if [ "X${rep_index}" != "X" ]; then
            ${VSCTL} --may-exist add-port ${bridge} ${port}  -- set interface ${port} type=dpdk ofport_request=${of_port_index} \
              -- set interface ${port} mtu_request=${MTU_INTERFACES} -- set bridge ${bridge} external_ids:ofport_index=$((${of_port_index}+1))
        else
            ${VSCTL} --may-exist add-port ${bridge} ${port}  -- set interface ${port} type=dpdk ofport_request=${of_port_index} \
              -- set interface ${port} mtu_request=${MTU_INTERFACES} -- set bridge ${bridge} external_ids:ofport_index=$((${of_port_index}+1))
        fi
    fi
    of_port_index=$((${of_port_index}+1))
}

# Add_patch_flows() adds openflow rules to set up bidirectional flows between the given ports and bridge.
# The flows uses `ofport_request` index numbers instead of port names.
# These are obtained from the interface table.

add_patch_flow(){
    local port1=$1
    local port2=$2
    local bridge=${3:-br-hbn}

    check_br ${bridge}

    msg "Setting flows for bridge: ${bridge} with ports: ${port1} and ${port2}"
    local port1_idx
    local port2_idx

    port1_idx=$(${VSCTL} list interface ${port1} 2>/dev/null | grep ofport_request | cut -d ":" -f2 | xargs)
    port2_idx=$(${VSCTL} list interface ${port2} 2>/dev/null | grep ofport_request | cut -d ":" -f2 | xargs)
    port1_idx_actual=$(${VSCTL} list interface ${port1} 2>/dev/null | grep 'ofport ' | cut -d ":" -f2 | xargs)
    port2_idx_actual=$(${VSCTL} list interface ${port2} 2>/dev/null | grep 'ofport ' | cut -d ":" -f2 | xargs)

    if [ "${port1_idx_actual}" != "-1" ] && [ "${port1_idx}"  != "${port1_idx_actual}" ]; then
      die "Port: ${port1} does not have ofport:${port1_idx} the same as ofport_request:${port1_idx_actual}"
    fi
    if [ "${port2_idx_actual}" != "-1" ] && [ "${port2_idx}"  != "${port2_idx_actual}" ]; then
      die "Port: ${port2} does not have ofport:${port2_idx} the same as ofport_request:${port2_idx_actual}"
    fi

    ${OFCTL} add-flow ${bridge} table=0,priority=0,in_port=${port1_idx},actions=output:${port2_idx}
    ${OFCTL} add-flow ${bridge} table=0,priority=0,in_port=${port2_idx},actions=output:${port1_idx}
}

add_peers(){
    local port_sfc=${1}
    local port_hbn=${2}
    local hbn_netdev=${3}
    local hbn_rep_ofport=${4}
    local bridge=${5}
    local sfc_name=p-${port_hbn}-sfc
    local hbn_name=p-${port_hbn}-hbn

    # The OF port index is increased atomically after the ovs-vsctl commands
    local index=$(${VSCTL} list interface ${sfc_name} 2>/dev/null | grep ofport_request | cut -d ":" -f2 | xargs)
    if [ -z "${index}" ]; then
      ${VSCTL} --may-exist add-port br-sfc ${sfc_name} -- \
        set int ${sfc_name} type=patch options:peer=${hbn_name} ofport_request=${SFC_OF_PORT_INDEX} -- \
        set bridge br-sfc external_ids:ofport_index=$(expr $SFC_OF_PORT_INDEX + 1)
    fi

    index=$(${VSCTL} list interface ${hbn_name} 2>/dev/null | grep ofport_request | cut -d ":" -f2 | xargs)
    if [ -z "${index}" ]; then
      ${VSCTL} --may-exist add-port br-hbn ${hbn_name} \
        -- set int ${hbn_name} type=patch options:peer=${sfc_name} ofport_request=${HBN_OF_PORT_INDEX} \
        -- set bridge br-hbn external_ids:ofport_index=$(expr $HBN_OF_PORT_INDEX + 1)
    fi

    # Set external ids on the patch port on br-hbn
    ${VSCTL} set interface ${hbn_name} external_ids:hbn_netdev="${hbn_netdev}"
    ${VSCTL} set interface ${hbn_name} external_ids:hbn_rep_ofport="${hbn_rep_ofport}"

    SFC_OF_PORT_INDEX=$(expr $SFC_OF_PORT_INDEX + 1)
    HBN_OF_PORT_INDEX=$(expr $HBN_OF_PORT_INDEX + 1)
    if ! [ -z "${ENABLE_BR_SFC_DEFAULT_FLOWS}" ] && [ "x${bridge}" != "xPatch" ]; then
      add_patch_flow ${port_sfc} ${sfc_name} br-sfc
    fi
    add_patch_flow ${hbn_name} ${port_hbn} br-hbn
}

set_patch(){

    local port1=$1
    local port2=$2
    local bridge=${3:-br-hbn}
    local port1_hbn_dev=$4
    local port1_of_port=$5
    local port2_hbn_dev=$6
    local port3_of_port=$7

    # Second port is expected to be always on the second port
    add_port ${port2} ${BR_HBN_NAME} HBN_OF_PORT_INDEX

    if [ "x${bridge}" == "x${BR_HBN_NAME}" ]; then
      add_port ${port1} ${bridge} HBN_OF_PORT_INDEX
      add_patch_flow ${port1} ${port2} ${bridge}

      ${VSCTL} set interface ${port1} external_ids:hbn_netdev="${port1_hbn_dev}"
      ${VSCTL} set interface ${port1} external_ids:hbn_rep_ofport="${port1_of_port}"
    elif [ "x${bridge}" == "xPatch" ]; then
      add_peers ${port1} ${port2} ${port1_hbn_dev} ${port1_of_port} ${bridge}
    else
      add_port ${port1} ${bridge} SFC_OF_PORT_INDEX
      add_peers ${port1} ${port2} ${port1_hbn_dev} ${port1_of_port} ${bridge}
    fi    
}


ensure_cni_bridge_iptables(){
  # Mgmt bridge configuration
  msg "Ensuring br-mgmt bridge"
  if ! (ip l | grep -q br-mgmt) ; then
        msg "Adding br-mgmt"
        ip link add br-mgmt type bridge
  fi
  msg "Ensuring br-mgmt to mgmt vrf"
  ip link set dev br-mgmt master mgmt
  ip link set dev br-mgmt up

    # Add CNI iptables rules to work with mgmt VRF
  msg "Ensuring IPtables rules"
  iptables -P FORWARD ACCEPT
  if ! iptables-save -t nat | grep -q ":CNI-HOSTPORT-DNAT"; then
    msg "Adding DNAT IPtables rules"
    iptables -t nat -N CNI-HOSTPORT-DNAT
    iptables -t nat -A PREROUTING -p tcp -m tcp --dport 8765 -j CNI-HOSTPORT-DNAT
  fi
}


#################################
###         Main              ###
#################################

source /etc/mellanox/sfc.conf  # Get mappings

initialize   # Initialize bridges, setup indexes, cleanup flows etc.

msg "Bridge name: ${BR_HBN_NAME}"
for MAP in "${MAPPINGS[@]}"; do
    string=($(echo "$MAP" | tr '~' '\n'))
    bridge=${string[0]}
    port1=${string[1]}
    port2=${string[2]}
    port1_hbn_dev=${string[3]}
    port1_of_port=${string[4]}
    port2_hbn_dev=${string[5]}
    port2_of_port=${string[6]}
    echo "Port mapping : ${bridge} <---> ${port1} <---> ${port2} <---> ${port1_hbn_dev} <---> ${port1_of_port} <---> ${port2_hbn_dev} <---> ${port2_of_port}"
    set_patch ${port1} ${port2} ${bridge} ${port1_hbn_dev} ${port1_of_port} ${port2_hbn_dev} ${port2_of_port}
done

for row in "${LINK_PROPAGATION_ARR[@]}"; do
    string=($(echo "$row" | tr ':' '\n'))
    interface=${string[0]}
    dependency=${string[1]}
    ${VSCTL} set interface ${interface} external_ids:dependencies="${dependency}"
done

# Store index for future.
${VSCTL} set bridge ${BR_HBN_NAME} external_ids:ofport_index=${HBN_OF_PORT_INDEX}
if [ -z "${BR_SFC_NAME}" ]; then
  ${VSCTL} set bridge ${BR_SFC_NAME} external_ids:ofport_index=${SFC_OF_PORT_INDEX}
fi

ensure_cni_bridge_iptables

msg "SFC Completed"
date -u >> /tmp/sfc-activated
